/*
 * smtp.c
 *
 * This file is part of msmtp, an SMTP client.
 *
 * Copyright (C) 2000, 2003, 2004
 * Martin Lambers <marlam@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   msmtp is released under the GPL with the additional exemption that
 *   compiling, linking, and/or using OpenSSL is allowed.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
extern int errno;

#ifdef USE_GSASL
#include <gsasl.h>
#else
#include "base64.h"
#include "md5_apps.h"
#endif

#include "list.h"
#include "net.h"
#ifdef HAVE_SSL
#include "tls.h"
#endif /* HAVE_SSL */
#include "xmalloc.h"
#include "smtp.h"


/* This defines the maximum number of lines in a multiline server reply.
 * This limit exists to prevent an extremely long reply from eating
 * all the memory.
 * The longest server reply we have to expect is the repsonse to the EHLO
 * command. We should have enough lines for every SMTP extension known today
 * plus more lines for future extensions.
 */
#define SMTP_MAXLINES 50

/* This defines the length of the *input* buffer for SMTP messages.
 * According to RFC 2821, SMTP commands and SMTP messages may contain
 * at most 512 characters including CRLF, thus the *minimum* size is
 * 512 characters.
 */
#define SMTP_BUFSIZE 1024

/* This defines the length of the *output* buffer for SMTP commands, without
 * CRLF. According to RFC 2821, SMTP commands may contain at most 512
 * characters including CRLF, thus the *maximum* size should be 510 characters.
 * Bus this is not sufficient for some authentication mechanisms, notably
 * GSSAPI, so this value must be higher.
 */
#define SMTP_MAXCMDLEN 1022

/* This is the buffer length for copying the mail to the SMTP server.
 * According to RFC 2822, a line in a mail can contain at most 998 
 * characters + CRLF. Plus one character for '\0' = 1001 characters.
 * All lines should fit in a buffer of this size.
 * However, this length is not a limit; smtp_copy_mail() will accept 
 * arbitrary long lines. 
 */
#define MAIL_BUFSIZE 1024


/*
 * smtp_new()
 *
 * see smtp.h
 */

smtp_server_t smtp_new(FILE *debug)
{
    smtp_server_t srv;

    srv.fd = -1;
#ifdef HAVE_SSL
    tls_clear(&srv.tls);
#endif /* HAVE_SSL */
    srv.cap.flags = 0;
    srv.cap.size = 0;
    srv.debug = debug;
    return srv;
}


/*
 * smtp_connect()
 *
 * see smtp.h
 */

int smtp_connect(smtp_server_t *srv, const char *host, int port, char *errstr)
{
    return net_open_socket(host, port, &srv->fd, errstr);
}


/*
 * smtp_get_msg()
 * 
 * This function gets a message from the SMTP server 'srv'.
 * In case of success, 'msg' will contain a pointer to a newly created list, 
 * and each member of this list will contain one line of the message as an 
 * allocated string. The list will contain at least one line. Each line will 
 * be at least 4 characters long: The three digit status code plus a ' ' or '-'.
 * Each line will be at most SMTP_BUFSIZE characters long, including '\0'.
 * The return code will be EOK.
 * In case of failure, 'msg' will be NULL, and one of the following error codes
 * will be returned: SMTP_EIO, SMTP_EPROTO
 */

int smtp_get_msg(smtp_server_t *srv, list_t **msg, char *errstr)
{
    list_t *l;
    list_t *lp;
    char line[SMTP_BUFSIZE];
    int counter;
    size_t len;

    *msg = NULL;
    l = list_new();
    lp = l;

    counter = 0;
    do
    {
#ifdef HAVE_SSL
	if (tls_is_active(&srv->tls))
	{
	    if (tls_gets(&srv->tls, line, SMTP_BUFSIZE, errstr) != TLS_EOK)
	    {
		list_xfree(l, free);
		return SMTP_EIO;
	    }
	}
	else
	{
#endif /* HAVE_SSL */
	    if (net_gets(srv->fd, line, SMTP_BUFSIZE, errstr) != NET_EOK)
	    {
		list_xfree(l, free);
		return SMTP_EIO;
	    }
#ifdef HAVE_SSL
	}
#endif /* HAVE_SSL */
	len = strlen(line);
	if (len < 4 
		|| !(isdigit((unsigned char)line[0]) 
		    && isdigit((unsigned char)line[1]) 
		    && isdigit((unsigned char)line[2]) 
		    && (line[3] == ' ' || line[3] == '-'))
		|| line[len - 1] != '\n')
	{
	    list_xfree(l, free);
	    /* The string is not necessarily a reply (it may be the initial OK
	     * message), but this is the term used in the RFCs.
	     * An empty reply is a special case of an invalid reply - this
	     * differentiation may help the user. */
	    snprintf(errstr, errstr_bufsize, "SMTP server sent an %s reply",
		    (counter == 0 && len == 0) ? "empty" : "invalid");
	    return SMTP_EPROTO;
	}
	/* kill CRLF */
	line[len - 1] = '\0';
	if (line[len - 2] == '\r')
	{
	    line[len - 2] = '\0';
	}
	list_insert(lp, xstrdup(line));
	counter++;
	lp = lp->next;
	if (srv->debug)
	{
	    fprintf(srv->debug, "<-- %s\n", line);
	}
    }
    while (line[3] == '-' && counter <= SMTP_MAXLINES);

    if (counter > SMTP_MAXLINES)
    {
	list_xfree(l, free);
	snprintf(errstr, errstr_bufsize, 
		"Rejecting SMTP server reply that is longer than %d lines. "
		"Increase SMTP_MAXLINES.", SMTP_MAXLINES);
	return SMTP_EPROTO;
    }

    *msg = l;
    return SMTP_EOK;
}


/*
 * smtp_msg_status()
 *
 * see smtp.h
 */

int smtp_msg_status(list_t *msg)
{
    return atoi(msg->next->data);
}


/*
 * smtp_put()
 *
 * This function writes a string to the SMTP server 'srv'. If 'newline' is set,
 * it will write TCP CRLF ('\r\n') after the string.
 * Used error codes: SMTP_EIO
 */

int smtp_put(smtp_server_t *srv, char *s, int newline, char *errstr)
{
    int e = 0;

#ifdef HAVE_SSL
    if (tls_is_active(&srv->tls))
    {
	if (tls_puts(&srv->tls, s, errstr) == TLS_EOK && newline)
	{
	    e = (tls_puts(&srv->tls, "\r\n", errstr) != TLS_EOK);
	}
    }
    else
    {
#endif /* HAVE_SSL */
	if (net_puts(srv->fd, s, errstr) == NET_EOK && newline)
	{
	    e = (net_puts(srv->fd, "\r\n", errstr) != NET_EOK);
	}
#ifdef HAVE_SSL
    }
#endif /* HAVE_SSL */
    if (e)
    {
	return SMTP_EIO;
    }
    if (srv->debug)
    {
	fprintf(srv->debug, "--> %s\n", s);
    }
    
    return SMTP_EOK;
}


/*
 * smtp_send_cmd()
 *
 * This function writes a string to the SMTP server 'srv'. The string may not 
 * be longer than SMTP_MAXCMDLEN characters (see above). TCP CRLF ('\r\n') will be 
 * appended to the string. Use this function to send SMTP commands (not mail data)
 * to the SMTP server.
 * Used error codes: SMTP_EIO, SMTP_EINVAL
 */

int smtp_send_cmd(smtp_server_t *srv, char *errstr, const char *format, ...)
{
    char line[SMTP_MAXCMDLEN + 1];
    int count;
    va_list args;

    va_start(args, format);
    count = vsnprintf(line, SMTP_MAXCMDLEN + 1, format, args);	
    va_end(args);
    if (count >= SMTP_MAXCMDLEN + 1)
    {
	snprintf(errstr, errstr_bufsize, 
		"Cannot send SMTP command because it is "
		"longer than %d characters. Increase SMTP_MAXCMDLEN.", 
		SMTP_MAXCMDLEN);
	return SMTP_EINVAL;
    }
    return smtp_put(srv, line, 1, errstr);
}


/*
 * smtp_get_greeting()
 *
 * see smtp.h
 */

int smtp_get_greeting(smtp_server_t *srv, list_t **errmsg, char **buf, char *errstr)
{
    int e;
    list_t *msg;
    
    *errmsg = NULL;
    if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
    {
	return e;
    }
    if (smtp_msg_status(msg) != 220)
    {
	*errmsg = msg;
	snprintf(errstr, errstr_bufsize, 
		"Cannot get initial OK message from SMTP server");	
	return SMTP_EPROTO;
    }
    if (buf)
    {
	*buf = xmalloc((strlen((char *)msg->next->data + 4) + 1) * sizeof(char));
	strcpy(*buf, (char *)(msg->next->data) + 4);
    }
    list_xfree(msg, free);
    
    return SMTP_EOK;
}


/*
 * smtp_init()
 *
 * see smtp.h
 */

int smtp_init(smtp_server_t *srv, const char *ehlo_domain, list_t **errmsg, char *errstr)
{
    int e;
    list_t *ehlo_response;
    list_t *lp;
    char *s;
    size_t len;
    int i;
    
    srv->cap.flags = 0;
    srv->cap.size = 0;

    *errmsg = NULL;
    if ((e = smtp_send_cmd(srv, errstr, "EHLO %s", ehlo_domain)) != SMTP_EOK)
    {
	return e;
    }
    if ((e = smtp_get_msg(srv, &ehlo_response, errstr)) != SMTP_EOK)
    {
	return e;
    }
    if (smtp_msg_status(ehlo_response) != 250)
    {
	/* fall back to HELO */
	list_xfree(ehlo_response, free);
	if ((e = smtp_send_cmd(srv, errstr, "HELO %s", ehlo_domain)) != SMTP_EOK)
	{
	    return e;
	}
	if ((e = smtp_get_msg(srv, &ehlo_response, errstr)) != SMTP_EOK)
	{
	    return e;
	}
	if (smtp_msg_status(ehlo_response) != 250)
	{
	    *errmsg = ehlo_response;
	    snprintf(errstr, errstr_bufsize,
		    "SMTP server does not accept EHLO or HELO commands");
	    return SMTP_EPROTO;
	}
	list_xfree(ehlo_response, free);
	/* srv->cap.flags is 0 */
	return SMTP_EOK;
    }

    lp = ehlo_response;
    while (!list_is_empty(lp))
    {
	lp = lp->next;
	s = lp->data;
	len = strlen(s);
	/* we know that len is >= 4 */
	/* make line uppercase */
    	for (i = 4; (size_t)i < len; i++)
	{
	    s[i] = toupper((unsigned char)s[i]);
	}
	/* search capabilities */
	if (strncmp(s + 4, "STARTTLS", 8) == 0)
	{
    	    srv->cap.flags |= SMTP_CAP_STARTTLS;
	}
	else if (strncmp(s + 4, "DSN", 3) == 0)
	{
	    srv->cap.flags |= SMTP_CAP_DSN;
	}
	else if (strncmp(s + 4, "PIPELINING", 10) == 0)
	{
	    srv->cap.flags |= SMTP_CAP_PIPELINING;
	}
	else if (strncmp(s + 4, "SIZE", 4) == 0)
	{
	    srv->cap.flags |= SMTP_CAP_SIZE;
	    srv->cap.size = atol(s + 8);
	}
     	else if (strncmp(s + 4, "AUTH ", 5) == 0)
	{
    	    srv->cap.flags |= SMTP_CAP_AUTH;
	    if (strstr(s + 9, "PLAIN"))
	    {
    		srv->cap.flags |= SMTP_CAP_AUTH_PLAIN;
	    }
	    if (strstr(s + 9, "CRAM-MD5"))
	    {
		srv->cap.flags |= SMTP_CAP_AUTH_CRAM_MD5;
	    }
	    if (strstr(s + 9, "DIGEST-MD5"))
	    {
		srv->cap.flags |= SMTP_CAP_AUTH_DIGEST_MD5;
	    }
	    if (strstr(s + 9, "GSSAPI"))
	    {
		srv->cap.flags |= SMTP_CAP_AUTH_GSSAPI;
	    }
	    if (strstr(s + 9, "LOGIN"))
	    {
		srv->cap.flags |= SMTP_CAP_AUTH_LOGIN;
	    }
	    if (strstr(s + 9, "NTLM"))
	    {
		srv->cap.flags |= SMTP_CAP_AUTH_NTLM;
	    }
	}
    }

    list_xfree(ehlo_response, free);
    return SMTP_EOK;
}


/*
 * smtp_tls_init()
 *
 * see smtp.h
 */

#ifdef HAVE_SSL
int smtp_tls_init(smtp_server_t *srv, const char *tls_key_file, 
	const char *tls_ca_file, const char *tls_trust_file, char *errstr)
{
    return tls_init(&srv->tls, tls_key_file, tls_ca_file, tls_trust_file, errstr);
}
#endif /* HAVE_SSL */


/*
 * smtp_tls_starttls()
 *
 * see smtp.h
 */

#ifdef HAVE_SSL
int smtp_tls_starttls(smtp_server_t *srv, list_t **error_msg, char *errstr)
{
    int e;
    list_t *msg;

    *error_msg = NULL;
    if ((e = smtp_send_cmd(srv, errstr, "STARTTLS")) != SMTP_EOK)
    {
	return e;
    }
    if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
    {
	return e;
    }
    if (smtp_msg_status(msg) != 220)
    {
	*error_msg = msg;
	snprintf(errstr, errstr_bufsize, 
		"SMTP server does not accept the STARTTLS command");
	return SMTP_EPROTO;
    }
    return SMTP_EOK;
}
#endif /* HAVE_SSL */


/*
 * smtp_tls()
 *
 * see smtp.h
 */

#ifdef HAVE_SSL
int smtp_tls(smtp_server_t *srv, const char *hostname, int tls_nocertcheck,
	char *errstr)
{
    return tls_start(&srv->tls, srv->fd, hostname, tls_nocertcheck, errstr);
}
#endif /* HAVE_SSL */


/*
 * smtp_auth_plain()
 * 
 * Do SMTP authentication via AUTH PLAIN.
 * The SMTP server must support SMTP_CAP_AUTH_PLAIN
 * Used error codes: SMTP_EIO, SMTP_EPROTO, SMTP_EAUTHFAIL, SMTP_EINVAL
 */

#ifndef USE_GSASL
int smtp_auth_plain(smtp_server_t *srv, const char *user, const char *password,
	list_t **error_msg, char *errstr)
{
    char *s;
    char *b64;
    size_t u_len;
    size_t p_len;
    list_t *msg;
    int e;
    int status;

    *error_msg = NULL;
    u_len = strlen(user);
    p_len = strlen(password);
    s = xmalloc((u_len + p_len + 3) * sizeof(char));
    s[0] = '\0';
    strcpy(s + 1, user);
    strcpy(s + u_len + 2, password);
    b64 = base64enc(s, 1 + u_len + 1 + p_len);
    free(s);
    
    if ((e = smtp_send_cmd(srv, errstr, "AUTH PLAIN %s", b64)) != SMTP_EOK)
    {
	free(b64);
	return e;
    }
    free(b64);
    if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
    {
	return e;
    }
    if ((status = smtp_msg_status(msg)) != 235)
    {
	*error_msg = msg;
	if (status == 504)
	{
	    snprintf(errstr, errstr_bufsize, 
		    "AUTH PLAIN not accepted by SMTP server");
	    return SMTP_EPROTO;
	}
	else
	{
	    snprintf(errstr, errstr_bufsize,
		    "SMTP server does not accept username/password "
		    "(AUTH PLAIN)");
	    return SMTP_EAUTHFAIL;
	}
    }
    list_xfree(msg, free);

    return SMTP_EOK;
}
#endif /* !USE_GSASL */


/*
 * smtp_auth_login()
 * 
 * Do SMTP authentication via AUTH LOGIN.
 * The SMTP server must support SMTP_CAP_AUTH_LOGIN
 * Used error codes: SMTP_EIO, SMTP_EPROTO, SMTP_EAUTHFAIL, SMTP_EINVAL
 */

#ifndef USE_GSASL
int smtp_auth_login(smtp_server_t *srv, const char *user, const char *password,
	list_t **error_msg, char *errstr)
{
    int e;
    list_t *msg;
    char *b64;
    
    *error_msg = NULL;
    if ((e = smtp_send_cmd(srv, errstr, "AUTH LOGIN")) != SMTP_EOK)
    {
	return e;
    }
    if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
    {
	return e;
    }
    if (smtp_msg_status(msg) != 334)
    {
	*error_msg = msg;
	snprintf(errstr, errstr_bufsize, 
		"AUTH LOGIN not accepted by SMTP server");
	return SMTP_EPROTO;
    }
    list_xfree(msg, free);
    b64 = base64enc(user, strlen(user));
    if ((e = smtp_send_cmd(srv, errstr, "%s", b64)) != SMTP_EOK)
    {
	free(b64);
	return e;
    }
    free(b64);
    if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
    {
	return e;
    }
    if (smtp_msg_status(msg) != 334)
    {
	*error_msg = msg;
	snprintf(errstr, errstr_bufsize,
		"SMTP server does not accept username (AUTH LOGIN)");
	return SMTP_EAUTHFAIL;
    }
    list_xfree(msg, free);
    b64 = base64enc(password, strlen(password));
    if ((e = smtp_send_cmd(srv, errstr, "%s", b64)) != SMTP_EOK)
    {
	free(b64);
	return e;
    }
    free(b64);
    if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
    {
	return e;
    }
    if (smtp_msg_status(msg) != 235)
    {
	*error_msg = msg;
	snprintf(errstr, errstr_bufsize, 
		"SMTP server does not accept username/password (AUTH LOGIN)");
	return SMTP_EAUTHFAIL;
    }
    list_xfree(msg, free);

    return SMTP_EOK;
}
#endif /* !USE_GSASL */


/*
 * smtp_auth_cram_md5()
 *
 * Do SMTP authentication via AUTH CRAM-MD5.
 * The SMTP server must support SMTP_CAP_AUTH_CRAM_MD5
 * Used error codes: SMTP_EIO, SMTP_EPROTO, SMTP_EAUTHFAIL, SMTP_ELIBFAILED, 
 * SMTP_EINVAL
 */

#ifndef USE_GSASL
int smtp_auth_cram_md5(smtp_server_t *srv, const char *user, const char *password,
	list_t **error_msg, char *errstr)
{
    unsigned char digest[16];
    char hex[] = "0123456789abcdef";
    char *b64;
    char *buf;
    char *p;
    size_t len;
    int i;
    list_t *msg;
    int e;
    
    *error_msg = NULL;
    if ((e = smtp_send_cmd(srv, errstr, "AUTH CRAM-MD5")) != SMTP_EOK)
    {
	return e;
    }
    if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
    {
	return e;
    }
    if (smtp_msg_status(msg) != 334)
    {
	*error_msg = msg;
	snprintf(errstr, errstr_bufsize, 
		"AUTH CRAM-MD5 not accepted by SMTP server");
	return SMTP_EPROTO;
    }
    /* we know the line is at least 4 characters long */
    if (!(b64 = base64dec((char *)(msg->next->data) + 4, &len)))
    {
	list_xfree(msg, free);
	snprintf(errstr, errstr_bufsize,
		"AUTH CRAM-MD5: server sent invalid challenge");
	return SMTP_EPROTO;
    }
    list_xfree(msg, free);
    if ((e = md5_hmac(password, strlen(password), b64, len, digest, 
		    errstr)) != MD5_EOK)
    {
	free(b64);
	return SMTP_ELIBFAILED;
    }
    free(b64);
    
    /* construct username + ' ' + digest_in_hex */
    len = strlen(user);
    buf = xmalloc((len + 1 + 32 + 1) * sizeof(char));
    strcpy(buf, user);
    p = buf + len;
    *p++ = ' ';
    for (i = 0; i < 16; i++)
    {
	p[2 * i] = hex[digest[i] >> 4];
	p[2 * i + 1] = hex[digest[i] & 0x0f];
    }
    p[32] = '\0';
    
    b64 = base64enc(buf, len + 33);
    free(buf);
    if ((e = smtp_send_cmd(srv, errstr, "%s", b64)) != SMTP_EOK)
    {
	free(b64);
	return e;
    }
    free(b64);
    if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
    {
	return e;
    }
    if (smtp_msg_status(msg) != 235)
    {
	*error_msg = msg;
	snprintf(errstr, errstr_bufsize, 
		"SMTP server does not accept username/password "
		"(AUTH CRAM-MD5)");
	return SMTP_EAUTHFAIL;
    }
    list_xfree(msg, free);

    return SMTP_EOK;
}
#endif /* !USE_GSASL */


/*
 * smtp_server_supports_authmech()
 *
 * see smtp.h
 */

int smtp_server_supports_authmech(smtp_server_t *srv, const char *mech)
{
    return ((strcmp(mech, "PLAIN") == 0 && 
		(srv->cap.flags & SMTP_CAP_AUTH_PLAIN))
	    || (strcmp(mech, "CRAM-MD5") == 0 && 
		(srv->cap.flags & SMTP_CAP_AUTH_CRAM_MD5))
	    || (strcmp(mech, "DIGEST-MD5") == 0 && 
		(srv->cap.flags & SMTP_CAP_AUTH_DIGEST_MD5))
	    || (strcmp(mech, "GSSAPI") == 0 && 
		(srv->cap.flags & SMTP_CAP_AUTH_GSSAPI))
	    || (strcmp(mech, "LOGIN") == 0 && 
		(srv->cap.flags & SMTP_CAP_AUTH_LOGIN))
	    || (strcmp(mech, "NTLM") == 0 && 
		(srv->cap.flags & SMTP_CAP_AUTH_NTLM)));
}


/*
 * smtp_client_supports_authmech()
 *
 * see smtp.h
 */

int smtp_client_supports_authmech(const char *mech)
{
#ifdef USE_GSASL

    int supported = 0;
    Gsasl *ctx;
    
    if (gsasl_init(&ctx) != GSASL_OK)
    {
	return 0;
    }
    supported = gsasl_client_support_p(ctx, mech);
    gsasl_done(ctx);
    return supported;
    
#else /* not USE_GSASL */
    
    if (strcmp(mech, "CRAM-MD5") == 0
	    || strcmp(mech, "PLAIN") == 0
	    || strcmp(mech, "LOGIN") == 0)
    {
	return 1;
    }
    else /* DIGEST-MD5, GSSAPI or NTLM */
    {
	return 0;
    }
    
#endif /* not USE_GSASL */
}


/*
 * smtp_auth()
 *
 * see smtp.h
 */

int smtp_auth(smtp_server_t *srv,
	const char *hostname,
	const char *user, 
	const char *password,
	const char *ntlmdomain,
	const char *auth_mech,
	list_t **error_msg,
	char *errstr)
{
#ifdef USE_GSASL
    int e;
    list_t *msg;
    Gsasl *ctx;
    Gsasl_session *sctx;
    char *input;
    char inbuf[SMTP_BUFSIZE];
    char *outbuf;
    int error_code;
    int auth_plain_special;

    *error_msg = NULL;
    if (strcmp(auth_mech, "") != 0 && 
	    !smtp_server_supports_authmech(srv, auth_mech))
    {
	snprintf(errstr, errstr_bufsize, 
		"SMTP server does not support authentication mechanism %s", 
		auth_mech);
	return SMTP_EUNAVAIL;
    }
    if ((error_code = gsasl_init(&ctx)) != GSASL_OK)
    {
	snprintf(errstr, errstr_bufsize, "GNU SASL: %s", 
		gsasl_strerror(error_code));
	return SMTP_ELIBFAILED;
    }
    if (strcmp(auth_mech, "") != 0 && !gsasl_client_support_p(ctx, auth_mech))
    {
	gsasl_done(ctx);
	snprintf(errstr, errstr_bufsize, 
		"GNU SASL: authentication mechanism %s not supported", 
		auth_mech);
	return SMTP_ELIBFAILED;
    }
    if (strcmp(auth_mech, "") == 0)
    {
	/* Choose "best" authentication mechanism. */
	/* TODO: use gsasl_client_suggest_mechanism()? */
	if (gsasl_client_support_p(ctx, "GSSAPI") 
		&& (srv->cap.flags & SMTP_CAP_AUTH_GSSAPI))
	{
	    auth_mech = "GSSAPI";
	}
	else if (gsasl_client_support_p(ctx, "DIGEST-MD5") 
		&& (srv->cap.flags & SMTP_CAP_AUTH_DIGEST_MD5))
	{
	    auth_mech = "DIGEST-MD5";
	}
	else if (gsasl_client_support_p(ctx, "CRAM-MD5") 
		&& (srv->cap.flags & SMTP_CAP_AUTH_CRAM_MD5))
	{
	    auth_mech = "CRAM-MD5";
	}
	else if (gsasl_client_support_p(ctx, "NTLM") 
		&& (srv->cap.flags & SMTP_CAP_AUTH_NTLM))
	{
	    auth_mech = "NTLM";
	}
#ifdef HAVE_SSL
	else if (tls_is_active(&srv->tls))
	{
	    if (gsasl_client_support_p(ctx, "PLAIN") 
		    && (srv->cap.flags & SMTP_CAP_AUTH_PLAIN))
	    {
		auth_mech = "PLAIN";
	    }
	    else if (gsasl_client_support_p(ctx, "LOGIN") 
		    && (srv->cap.flags & SMTP_CAP_AUTH_LOGIN))
	    {
		auth_mech = "LOGIN";
	    }
	}
#endif /* HAVE_SSL */
    }
    if (strcmp(auth_mech, "") == 0)
    {
	gsasl_done(ctx);
	snprintf(errstr, errstr_bufsize, 
		"GNU SASL: cannot agree on authentication mechanism");
	return SMTP_EUNAVAIL;
    }
    
    if ((error_code = gsasl_client_start(ctx, auth_mech, &sctx)) != GSASL_OK)
    {
	gsasl_done(ctx);
	snprintf(errstr, errstr_bufsize, 
		"GNU SASL: %s", gsasl_strerror(error_code));
	return SMTP_ELIBFAILED;
    }

    /* Set the authentication properties */
    if (user)
    {
	gsasl_property_set(sctx, GSASL_AUTHID, user);
	gsasl_property_set(sctx, GSASL_AUTHZID, user);
    }
    if (password)
    {
	gsasl_property_set(sctx, GSASL_PASSWORD, password);
    }
    /* For DIGEST-MD5 and GSSAPI */
    gsasl_property_set(sctx, GSASL_SERVICE, "smtp");
    if (hostname)
    {
	gsasl_property_set(sctx, GSASL_HOSTNAME, hostname);
    }
    /* For NTLM. Postfix does not care, MS IIS needs an arbitrary non-empty
     * string. */
    if (ntlmdomain)
    {
	gsasl_property_set(sctx, GSASL_REALM, ntlmdomain);
    }

    /* Bigg authentication loop */
    input = NULL;
    do
    {
	error_code = gsasl_step64(sctx, input, &outbuf);
	if (error_code != GSASL_OK && error_code != GSASL_NEEDS_MORE)
	{
	    gsasl_finish(sctx);
	    gsasl_done(ctx);
	    snprintf(errstr, errstr_bufsize, 
		    "GNU SASL: %s", gsasl_strerror(error_code));
	    return SMTP_ELIBFAILED;
	}
	if (!input)
	{
	    if (strcmp(auth_mech, "PLAIN") == 0 && outbuf[0])
	    {
		/* AUTH PLAIN needs special treatment because it needs to send the
		 * authentication data together with the AUTH PLAIN command */
		auth_plain_special = 1;
		if ((e = smtp_send_cmd(srv, errstr, 
				"AUTH PLAIN %s", outbuf)) != SMTP_EOK)
		{
		    gsasl_finish(sctx);
		    gsasl_done(ctx);
		    free(outbuf);
		    return e;
		}
	    }
	    else
	    {	    
		auth_plain_special = 0;
		if ((e = smtp_send_cmd(srv, errstr, 
				"AUTH %s", auth_mech)) != SMTP_EOK)
		{
		    gsasl_finish(sctx);
		    gsasl_done(ctx);
		    free(outbuf);
		    return e;
		}
	    }
	    if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
	    {
		gsasl_finish(sctx);
		gsasl_done(ctx);
	    	free(outbuf);
		return e;
	    }
	    if (smtp_msg_status(msg) != 334 && smtp_msg_status(msg) != 235)
	    {
		*error_msg = msg;
		gsasl_finish(sctx);
		gsasl_done(ctx);
		free(outbuf);
		snprintf(errstr, errstr_bufsize, 
			"GNU SASL, mechanism %s: authentication failed", auth_mech);
		return SMTP_EAUTHFAIL;
	    }	    
	    /* msg->next->data cannot be longer than SMTP_BUFSIZE-1 */
	    strcpy(inbuf, msg->next->data);
	    list_xfree(msg, free);
	    input = inbuf + 4;
	    if (auth_plain_special)
	    {
		free(outbuf);
		continue;
	    }
	}
	/* For all mechanisms except GSSAPI, testing for (outbuf[0]) works.
	 * GSSAPI needs an additional step with empty output. */
	if (outbuf[0] || (GSASL_NEEDS_MORE && (strcmp(auth_mech, "GSSAPI") == 0)))
	{
	    if ((e = smtp_send_cmd(srv, errstr, "%s", outbuf)) != SMTP_EOK)
    	    {
		gsasl_finish(sctx);
		gsasl_done(ctx);
		free(outbuf);
		return e;
	    }
	    if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
	    {
		gsasl_finish(sctx);
		gsasl_done(ctx);
		free(outbuf);
		return e;
	    }
	    if (smtp_msg_status(msg) != 334 && smtp_msg_status(msg) != 235)
	    {
		*error_msg = msg;
		gsasl_finish(sctx);
		gsasl_done(ctx);
		free(outbuf);
		snprintf(errstr, errstr_bufsize, 
			"GNU SASL, mechanism %s: authentication failed", auth_mech);
		return SMTP_EAUTHFAIL;
	    }	    
	    /* msg->next->data cannot be longer than SMTP_BUFSIZE-1 */
	    strcpy(inbuf, msg->next->data);
	    list_xfree(msg, free);
	    input = inbuf + 4;
	}
	free(outbuf);
    }
    while (error_code == GSASL_NEEDS_MORE);
    if (error_code != GSASL_OK)
    {
	gsasl_finish(sctx);
	gsasl_done(ctx);
	snprintf(errstr, errstr_bufsize, 
		"GNU SASL, mechanism %s: authentication failed: %s", 
		auth_mech, gsasl_strerror(error_code));
	return SMTP_EAUTHFAIL;
    }
    gsasl_finish(sctx);
    gsasl_done(ctx);
    /* For DIGEST-MD5, we need to send an empty answer to the last 334 
     * response before we get 235. */
    if (strncmp(inbuf, "235 ", 4) != 0)
    {
	if ((e = smtp_send_cmd(srv, errstr, "")) != SMTP_EOK)
	{
	    return e;
	}
	if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
	{
	    return e;
	}
	if (smtp_msg_status(msg) != 235)
	{
	    *error_msg = msg;
	    snprintf(errstr, errstr_bufsize, 
		    "GNU SASL, mechanism %s: authentication failed", auth_mech);
	    return SMTP_EAUTHFAIL;
	}
	list_xfree(msg, free);
    }
    return SMTP_EOK;

#else /* not USE_GSASL */
    
    *error_msg = NULL;
    if (strcmp(auth_mech, "") != 0 && 
	    !smtp_server_supports_authmech(srv, auth_mech))
    {
	snprintf(errstr, errstr_bufsize, 
		"SMTP server does not support authentication mechanism %s", 
		auth_mech);
	return SMTP_EUNAVAIL;
    }
    if (strcmp(auth_mech, "") == 0)
    {
	/* Choose "best" authentication mechanism. */
	if (srv->cap.flags & SMTP_CAP_AUTH_CRAM_MD5)
	{
	    auth_mech = "CRAM-MD5";
	}
#ifdef HAVE_SSL
	else if (tls_is_active(&srv->tls))
	{
	    if (srv->cap.flags & SMTP_CAP_AUTH_PLAIN)
	    {
		auth_mech = "PLAIN";
	    }
	    else if (srv->cap.flags & SMTP_CAP_AUTH_LOGIN)
	    {
		auth_mech = "LOGIN";
	    }
	}
#endif /* HAVE_SSL */
    }
    if (strcmp(auth_mech, "") == 0)
    {
	snprintf(errstr, errstr_bufsize, 
		"cannot agree on an authentication mechanism");
	return SMTP_EUNAVAIL;
    }

    /* CRAMD-MD5, PLAIN, LOGIN all need a user name and password */
    if (!user)
    {
	snprintf(errstr, errstr_bufsize,
		"authentication mechanism %s needs a user name", auth_mech);
	return SMTP_EUNAVAIL;
    }
    if (!password)
    {
	snprintf(errstr, errstr_bufsize,
		"authentication mechanism %s needs a password", auth_mech);
	return SMTP_EUNAVAIL;
    }

    if (strcmp(auth_mech, "CRAM-MD5") == 0)
    {
	return smtp_auth_cram_md5(srv, user, password, error_msg, errstr);
    }
    else if (strcmp(auth_mech, "PLAIN") == 0)
    {
	return smtp_auth_plain(srv, user, password, error_msg, errstr);
    }
    else if (strcmp(auth_mech, "LOGIN") == 0)
    {
	return smtp_auth_login(srv, user, password, error_msg, errstr);
    }
    else
    {
	snprintf(errstr, errstr_bufsize, 
		"authentication mechanism %s not supported", auth_mech);
	return SMTP_ELIBFAILED;
    }

#endif /* not USE_GSASL */
}


/*
 * smtp_send_envelope()
 * 
 * see smtp.h
 */

int smtp_send_envelope(smtp_server_t *srv,
	const char *envelope_from, 
	list_t *recipients,
	const char *dsn_notify,
	const char *dsn_return,
	list_t **error_msg,
	char *errstr)
{
    int e;
    int e_tmp;
    list_t *msg;
    list_t *lp;
    
    *error_msg = NULL;
    /* tell the server who wants to send mail */
    if (dsn_return)
    {
	e = smtp_send_cmd(srv, errstr, "MAIL FROM:<%s> RET=%s", envelope_from, dsn_return);
    }
    else
    {
	e = smtp_send_cmd(srv, errstr, "MAIL FROM:<%s>", envelope_from);
    }
    if (e != SMTP_EOK)
    {
	return e;
    }
    if (!(srv->cap.flags & SMTP_CAP_PIPELINING))
    {
	if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
	{
	    return e;
	}
	if (smtp_msg_status(msg) != 250)
	{
	    *error_msg = msg;
	    snprintf(errstr, errstr_bufsize, 
		    "envelope from address %s not accepted by SMTP server", 
		    envelope_from);
	    return SMTP_EINVAL;
	}
	list_xfree(msg, free);
    }
    /* tell the server who to send the mail to */
    lp = recipients;
    while (!list_is_empty(lp))
    {
	lp = lp->next;
	if (dsn_notify)
	{
	    e = smtp_send_cmd(srv, errstr, "RCPT TO:<%s> NOTIFY=%s", 
		    (char *)(lp->data), dsn_notify);
	}
	else
	{
	    e = smtp_send_cmd(srv, errstr, "RCPT TO:<%s>", (char *)(lp->data));
	}
	if (e != SMTP_EOK)
	{
	    return e;
	}
	if (!(srv->cap.flags & SMTP_CAP_PIPELINING))
	{
	    if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
	    {
		return e;
	    }
	    if (smtp_msg_status(msg) != 250)
	    {
		*error_msg = msg;
		snprintf(errstr, errstr_bufsize, 
			"recipient address %s not accepted by SMTP server", 
			(char *)(lp->data));
		return SMTP_EINVAL;
	    }
	    list_xfree(msg, free);
	}
    }
    /* get all pipelined replies if pipelining is used */
    if (srv->cap.flags & SMTP_CAP_PIPELINING)
    {
	if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
	{
	    return e;
	}
     	if (smtp_msg_status(msg) != 250)
	{
	    *error_msg = msg;
	    snprintf(errstr, errstr_bufsize,
		    "envelope from address %s not accepted by SMTP server", 
		    envelope_from);
	    e = SMTP_EINVAL;
	}
	else
	{
	    list_xfree(msg, free);
	}
	lp = recipients;
	while (!list_is_empty(lp))
	{
	    lp = lp->next;
	    if ((e_tmp = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
	    {
		return e_tmp;
	    }
    	    if (smtp_msg_status(msg) != 250)
	    {
		if (e == SMTP_EOK)
		{
		    *error_msg = msg;
		    e = SMTP_EINVAL; 
		    snprintf(errstr, errstr_bufsize, 
			    "recipient address %s not accepted by SMTP server", 
			    (char *)(lp->data));
		}
		else
		{
		    list_xfree(msg, free);
		}
	    }
	    else
	    {
		list_xfree(msg, free);
	    }
	}
	if (e != SMTP_EOK)
	{
	    return e;
	}
    }
    /* announce the mail */
    if ((e = smtp_send_cmd(srv, errstr, "DATA")) != SMTP_EOK)
    {
	return e;
    }
    if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
    {
	return e;
    }
    if (smtp_msg_status(msg) != 354)
    {
	*error_msg = msg;
	snprintf(errstr, errstr_bufsize, 
		"SMTP server does not accept the mail");
	return SMTP_EUNAVAIL;
    }
    list_xfree(msg, free);
    
    return SMTP_EOK;
}


/*
 * smtp_send_mail()
 *
 * see smtp.h
 */

int smtp_send_mail(smtp_server_t *srv, FILE *mailf, int keep_bcc, 
	long *mailsize, char *errstr)
{
    char buffer[MAIL_BUFSIZE];
    size_t len;
    int in_header;
    int in_bcc;
    int line_starts;
    int line_continues;
    int e;
    
    in_header = 1;
    in_bcc = 0;
    line_continues = 0;
    e = SMTP_EOK;
    while (fgets(buffer, (int)sizeof(buffer), mailf) != NULL)
    {
	line_starts = !line_continues;
	len = strlen(buffer);
	if (len > 0 && buffer[len - 1] == '\n')
	{
	    buffer[--len] = '\0';
	    line_continues = 0;
	}
	else if (len == MAIL_BUFSIZE - 1)
	{
	    line_continues = 1;
	}
	else
	{
	    /* last line lacks a newline character */
	    line_continues = 0;
	}
	if (!keep_bcc)
	{
	    if (line_starts && in_header && buffer[0] == '\0')
	    {
		in_header = 0;
	    }
	    if (in_header)
	    {
		if (line_starts)
		{
		    if (!in_bcc)
		    {
			if (strncmp(buffer, "Bcc: ", 5) == 0)
			{
			    in_bcc = 1;
			    /* remove Bcc header by ignoring this line */
			    continue;
			}
		    }
		    else
		    {
			/* continued header lines begin with "horizontal whitespace"
			 * (RFC 2822, section 2.2.3) */
			if (buffer[0] == '\t' || buffer[0] == ' ')
			{
			    /* remove Bcc header by ignoring this line */
			    continue;
			}
			else
			{
			    in_bcc = 0;
			}
		    }
		}
		else
		{
		    if (in_bcc)
		    {
			/* remove Bcc header by ignoring this line */
		    	continue;
		    }
		}
	    }
	}
	if (line_starts && buffer[0] == '.')
	{
	    e = smtp_put(srv, ".", 0, errstr);
	    /* do not count the quoting dot in 'mailsize' */
	}
	if (e == SMTP_EOK)
	{
    	    e = smtp_put(srv, buffer, !line_continues, errstr);
	    /* count a newline as one character */
	    *mailsize += (long)len + (line_continues ? 0 : 1);
	}
    	if (e != SMTP_EOK)
	{
	    return e;
	}
    }
    if (ferror(mailf))
    {
	snprintf(errstr, errstr_bufsize, 
		"input error while reading the mail");
	return SMTP_EIO;
    }

    return SMTP_EOK;
}


/*
 * smtp_end_mail()
 *
 * see smtp.h
 */

int smtp_end_mail(smtp_server_t *srv, list_t **error_msg, char *errstr)
{
    int e;
    list_t *msg;
    
    if ((e = smtp_send_cmd(srv, errstr, ".")) != SMTP_EOK)
    {
	return e;
    }
    if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
    {
	return e;
    }
    if (smtp_msg_status(msg) != 250)
    {
	*error_msg = msg;
	snprintf(errstr, errstr_bufsize, 
		"SMTP server did not accept the mail");
	return SMTP_EUNAVAIL;
    }
    list_xfree(msg, free);
    
    return SMTP_EOK;
}


/*
 * smtp_quit()
 *
 * see smtp.h
 */

int smtp_quit(smtp_server_t *srv, char *errstr)
{
    int e;
    list_t *msg;

    if ((e = smtp_send_cmd(srv, errstr, "QUIT")) == SMTP_EOK)
    {
	e = smtp_get_msg(srv, &msg, errstr);
    }
    if (msg)
    {
	list_xfree(msg, free);
    }
    return e;
}


/*
 * smtp_close()
 *
 * see smtp.h
 */

void smtp_close(smtp_server_t *srv)
{
#ifdef HAVE_SSL
    if (tls_is_active(&srv->tls))
    {
	tls_close(&srv->tls);
    }
#endif /* HAVE_SSL */
    net_close_socket(srv->fd);
}
